/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import nl.cloudfarming.client.geoviewer.RangedLegendPalette;
import nl.cloudfarming.client.isobus.model.IsobusGrid;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.loaders.FileEntry;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiDataObject.Entry;
import org.openide.loaders.MultiFileLoader;
import org.openide.util.Exceptions;

/**
 * File loader for Grid files Grid files are in the format
 * GRD[0-9][0-9][0-9][0-9][0-9].BIN or this format in lower case
 * 
 * Mandatory files : GRD[0-9][0-9][0-9][0-9][0-9].BIN or this format in lower case — grid format;
 * TASKDATA.XML or this format in lower case —the task data file which has references to the primary grid file;
 * 
 * Optional files : .TLG[0-9][0-9][0-9][0-9][0-9].BIN or this format in lower case — contains time log values; 
 *
 * @author Merijn Zengers
 */
public class GridFileLoader extends MultiFileLoader {
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.type.GridFileLoader");

    public static final String PRIMARY_PATTERN = "(?i)grd[0-9]{5}\\.bin";
    // secondary files and dependendies:
    private static final String TASKDATA_PATTERN = TaskDataFileLoader.PRIMARY_PATTERN;
    
    public GridFileLoader() {
        super(GridDataObject.class.getName());
    }

    @Override
    protected String actionsContext() {
        StringBuilder builder = new StringBuilder("Loaders/");
        builder.append(GridDataObject.MIME_TYPE);
        builder.append("/Actions/");
        return builder.toString();
    }

    /**
     * Finds the primary file, should only recognize itself.
     *
     * @param fo the fileObject to find the primary file by
     * @return
     */
    @Override
    protected FileObject findPrimaryFile(FileObject fo) {
        // never recognize folders.
        if (fo.isFolder()) {
            return null;
        }
        Pattern primaryPattern = Pattern.compile(PRIMARY_PATTERN);

        if (primaryPattern.matcher(fo.getNameExt()).find()) {
            return fo;
        }
        LOGGER.log(Level.FINEST, "No primary file found for FileObject {0}", fo);
        return null;
    }

    /**
     * Creates the MultidataObject for the file with format
     * GRD[0-9][0-9][0-9][0-9][0-9].BIN or toLower().
     * Creates no secondary files, depends on taskdata xml.
     *
     * @param primaryFile the Primary file to create the DataObject for
     * @return the created DataObject
     * @throws DataObjectExistsException if the DataObject already exists
     * @throws IOException
     */
    @Override
    protected MultiDataObject createMultiObject(FileObject primaryFile) throws DataObjectExistsException, IOException {

        List<FileObject> secondary = new ArrayList<>();
        Pattern taskDataPattern = Pattern.compile(TASKDATA_PATTERN);
        FileObject taskDataFile = null;
        InvalidDataObject.Detail invalidDataDetail; // userfriendly message to be used in tooltip or error panel.

        for (FileObject child : primaryFile.getParent().getChildren()) {
            if (!child.equals(primaryFile)) {
                if (taskDataPattern.matcher(child.getNameExt()).find()) {
                    // required to load data but don't claim it as a secondary file:
                    taskDataFile = child;
                }
            }
        }

        if (taskDataFile == null) {
            LOGGER.log(Level.SEVERE, "No taskdata.xml file found for {0}", primaryFile.getNameExt());
            invalidDataDetail = InvalidDataObject.detail(Bundle.InvalidDataObject_error_missing_dependency("taskdata.xml"));
        } else {
            
            try {
                DataObject dataObject = DataObject.find(taskDataFile);
                if (dataObject instanceof TaskDataObject) {
                    TaskDataObject isobusDataObject = (TaskDataObject) dataObject;
                    //FIXME: moving this step from DO to loader causes java.lang.StackOverflowError in unit tests...
                    IsobusGrid grid = isobusDataObject.getGrid(primaryFile);
                    IsobusValueGrid valueGrid = GridReader.getValues(primaryFile, grid, isobusDataObject.getTreatmentzones(primaryFile));
                    return new GridDataObject(primaryFile, this, secondary, grid, valueGrid);
                } else {
                    invalidDataDetail = InvalidDataObject.detail(Bundle.InvalidDataObject_error_invalid_dependency("taskdata.xml"));
                }         
            } catch (DataObjectNotFoundException ex) {
                Exceptions.attachSeverity(ex, Level.INFO);
                Exceptions.printStackTrace(ex);
                LOGGER.log(Level.SEVERE, "Exception while loading timelog binary data: {0}", ex.toString());
                invalidDataDetail = InvalidDataObject.detail("Exception while loading timelog binary data", ex);                
            }
            
        }

        return new InvalidDataObject(primaryFile, this, GridDataObject.ICON_BASE, invalidDataDetail);
    }

    /**
     * Creates a primary entry for the primary files
     *
     * @param obj the dataObject to create a primary entry for
     * @param primaryFile the primary file to add as primary entry
     * @return the created primary entry
     */
    @Override
    protected Entry createPrimaryEntry(MultiDataObject obj, FileObject primaryFile) {
        Entry entry = new FileEntry(obj, primaryFile);
        LOGGER.log(Level.FINEST, "Creating primary entry for {0}", primaryFile.getName());
        return entry;
    }

    /**
     * Creates a secondary entry for the provided MultiDataObject
     * @param obj the MultiDataObject to create the secondary entry for
     * @param secondaryFile the secondary file to create the entry from
     * @return The created entry
     */
    @Override
    protected Entry createSecondaryEntry(MultiDataObject obj, FileObject secondaryFile) {
//        Pattern taskDataPattern = Pattern.compile(TASKDATA_PATTERN);
//        assert taskDataPattern.matcher(secondaryFile.getNameExt()).find();
        assert false : "a grid file should not have secondary files";

        FileEntry entry = new FileEntry(obj, secondaryFile);
        LOGGER.log(Level.FINEST, "Creating secondary entry for {0}", secondaryFile.getName());
        return entry;
    }
}
