/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import java.io.IOException;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import nl.cloudfarming.client.isobus.model.BinaryReader;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.IsobusGrid;
import nl.cloudfarming.client.isobus.model.IsobusTreatmentZone;
import nl.cloudfarming.client.isobus.model.ProcessDataVariable;
import org.openide.filesystems.FileObject;
import org.openide.util.Exceptions;

/**
 * Class which reads binary grid files.
 * 
 * There are two types of files:
 * <ul>
 * <li>TYPE_1: the binary file is a lookup table which contains the treatmentZoneCode for every square in the grid.</li>
 * <li>TYPE_2: the binary file contains the values of the ProcessDataVariable.processDataValue
 * </ul>
 * 
 * @author Merijn Zengers
 * @author johan
 */
public class GridReader {

    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.model");
        
    private GridReader() {}

    /**
     * Creates a grid of raw process data values fro a binary file,
     * either by lookup via treatment zone codes from the binary (TYPE_1),
     * or by using the binary values directly (TYPE_2).
     * 
     * Take Notice!!: the values in the grid  are ordered from
     * right Top to left bottom inverted from the binary.
     * 
     * @param gridFile the binary GRD file
     * @param grid the grid element from the taskdata file
     * @param availableTreatmentZones the treatment zones that belong to the parent task of this grid.
     * @return a list of the read values
     */
    public static IsobusValueGrid getValues(FileObject gridFile, IsobusGrid grid, List<IsobusTreatmentZone> availableTreatmentZones){
        int numRows = new Long(grid.getMaximumRow()).intValue();
        int numColumns = new Long(grid.getMaximumColumn()).intValue();
        
        // Decrease rows and columns with one because IsobusTreatmentZoneGrid runs from 0 
        int rows = numRows - 1;
        int columns = 0;
        long value;
        
        DDI ddi = null;
        
        switch (grid.getType()) {
            case TYPE_1:
                ProcessDataVariable pdv = availableTreatmentZones.iterator().next().getPDV().get(0);
                ddi = DDI.findById(pdv.getProcessDataDDI());
                break;
            case TYPE_2:
                // find the prototype treatment zone via the grid-element reference:
                for (IsobusTreatmentZone treatmentZone : availableTreatmentZones) {
                    if (treatmentZone.getTreatmentZoneCode() == grid.getTreatmentZoneCode()) {
                        ddi = DDI.findById(treatmentZone.getPDV().get(0).getProcessDataDDI());
                        break;
                    }
                }
                break;
            default:
                throw new IllegalArgumentException("unsupported grid type: " + grid.getType());
        }
        
        IsobusValueGrid valueGrid = new IsobusValueGrid(numRows, numColumns, ddi);
        
        try (BinaryReader binaryReader = new BinaryReader(gridFile.getInputStream())){
                    
            while (binaryReader.available()) {
                
                switch (grid.getType()) {
                    case TYPE_1:
                        value = binaryReader.read(GridFieldEnum.TREATMENT_ZONE_CODE_TYPE);
                        boolean found = false;
                        for (IsobusTreatmentZone treatmentZone : availableTreatmentZones) {
                            if (treatmentZone.getTreatmentZoneCode() == value) {
                                valueGrid.insert(rows, columns, treatmentZone.getPDV().get(0).getProcessDataValue());
                                found = true;
                                //Use decrease instead of increase 
                                // Our raster starts at the left top and ends at right bottom
                                // The bytes provided from the grid file starts at the left bottom and ends at the right top
                                // It has to be inverted.
                                columns++;
                                if(columns == numColumns){
                                    rows--;
                                    columns = 0;
                                }
                            }
                        }
                        assert (found);
                        break;
                    case TYPE_2:
                        value = binaryReader.read(GridFieldEnum.TREATMENT_ZONE_VALUE_TYPE);
                        valueGrid.insert(rows, columns, (int)value);
                        //
                        columns++;
                        if(columns == numColumns){
                            rows--;
                            columns = 0;
                        }
                        break;
                }
            }
        
        } catch (NullPointerException | IOException ex) {
            LOGGER.log(Level.SEVERE, "Failed to read Treatmentzones from gridfile {0} with exception {1}", new Object[]{gridFile.getName(), ex});
            Exceptions.printStackTrace(ex);
        }        
        return valueGrid;
    }    

}
