/*
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *    johan - initial API and implementation and/or initial documentation
 */
package nl.cloudfarming.client.isobus.type;

import java.awt.Image;
import java.awt.datatransfer.Transferable;
import java.util.Objects;
import javax.swing.Action;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JToolBar;
import net.miginfocom.swing.MigLayout;
import nl.cloudfarming.client.util.swing.BodyPanel;
import org.netbeans.core.spi.multiview.CloseOperationState;
import org.netbeans.core.spi.multiview.MultiViewElement;
import org.netbeans.core.spi.multiview.MultiViewElementCallback;
import org.openide.awt.UndoRedo;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiFileLoader;
import org.openide.nodes.FilterNode;
import org.openide.nodes.Node;
import org.openide.util.ImageUtilities;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.datatransfer.PasteType;
import org.openide.windows.TopComponent;

/**
 * Placeholder data object for invalid files.
 * 
 * <p>Common use cases:</p>
 * <ul>
 * <li>invalid file contents (e.g. XML could not be parsed);
 * <li>missing/invalid mandatory secondary files or other dependencies;
 * </ul>
 * 
 * <p>Currently mostly useful for secondary files, as primary files are usually filtered by mime type or data object subclass.</p>
 * 
 * @author johan
 */
@NbBundle.Messages({
    // Bundle texts for common errors like missing/invalid dependencies (including mandatory secondary files):
    //TODO: could add factory methods to make these reusable messages easier accessible to other modules.
    "# {0} - dependency name",
    "InvalidDataObject error missing dependency=Cannot load file because {0} is missing",
    "# {0} - dependency name",
    "InvalidDataObject error invalid dependency=Cannot load file because {0} is invalid",
})
public class InvalidDataObject extends MultiDataObject {
    // TODO: move to util/util-swing module?; would add a list of extra openide/netbeans deps; could create util-nb instead.
    
    public static final String MIME_TYPE = "application/x-invalid";
    
    private final String originalIconBase;
    // originalMimeType. originalDOclass ?
    private final Detail detail;
    
    public InvalidDataObject(FileObject fo, MultiFileLoader loader, String originalIconBase, Detail detail) throws DataObjectExistsException {
        super(fo, loader);
        registerEditor(MIME_TYPE, true);
        this.detail = Objects.requireNonNull(detail);
        this.originalIconBase = originalIconBase;
    }

    @Override
    protected Node createNodeDelegate() {
        return new InvalidNode(super.createNodeDelegate());
    }

    public Detail getDetail() {
        return detail;
    }
    
    private static final String ERROR_ICON = "nl/cloudfarming/client/util/swing/error.png";
    private static final String ERROR_BADGE = "nl/cloudfarming/client/util/swing/error-badge.png";
            
    @NbBundle.Messages({
        "# {0} - original node name",
        "InvalidNode display name={0} (invalid)",
        "# {0} - original file name",
        "InvalidNode tooltip=File {0} was not loaded",
    })
    public class InvalidNode extends FilterNode {

        public InvalidNode(Node original) {
            super(original);
        }

        @Override
        public Image getIcon(int type) {
            // FIXME: can't use BadgeIcon yet as the Node API requires an Image.
            if (originalIconBase != null && !originalIconBase.isEmpty()) {
                Image o = ImageUtilities.loadImage(originalIconBase, true);
                Image b = ImageUtilities.loadImage(ERROR_BADGE, true);
                // offsets to show badge in bottom right corner:
                int x = o.getWidth(null)  - b.getWidth(null);
                int y = o.getHeight(null) - b.getHeight(null);
                return ImageUtilities.mergeImages(o, b, x, y);
            } else {
                return ImageUtilities.loadImage(ERROR_ICON);
            }
        }

        @Override
        public String getDisplayName() {
            return Bundle.InvalidNode_display_name(getOriginal().getName());
        }

        @Override
        public String getShortDescription() {
            // detail message might be too large for tooltip
            return Bundle.InvalidNode_tooltip(getPrimaryFile().getNameExt());
        }

        @Override
        public boolean canRename() {
            return false;
        }

        @Override
        public boolean canDestroy() {
            return true;
        }

        @Override
        public PasteType getDropType(Transferable t, int action, int index) {
            return null;
        }
    }
    
    public static Detail detail(String message) {
        return new  Detail(message, null);
    }
    public static Detail detail(String message, Throwable throwable) {
        return new Detail(message, throwable);
    }
    
    /**
     * Minimal information to show why loading failed or an object is considered invalid.
     * 
     * TODO: look into multiple messages per detail or multiple details per data object.
     */
    public static class Detail {
        private final String message;
        private final Throwable throwable;

        public Detail(String message, Throwable throwable) {
            this.message = message;
            this.throwable = throwable;
        }

        public String getMessage() {
            return message;
        }

        public Throwable getThrowable() {
            return throwable;
        }
    }
    
    /**
     * 
     */
    @MultiViewElement.Registration(
            displayName = "#LBL_InvalidData_VISUAL",
            iconBase = ERROR_ICON,
            mimeType = InvalidDataObject.MIME_TYPE,
            persistenceType = TopComponent.PERSISTENCE_NEVER,
            preferredID = "InvalidData",
            position = 1000)
    @NbBundle.Messages({
        "LBL_InvalidData_VISUAL=Invalid data",
    })
    public static class VisualElement implements MultiViewElement {
        private JToolBar toolbar = new JToolBar();
        private transient MultiViewElementCallback callback;
        private InvalidDataObject obj;
        private BodyPanel bodyPanel;

        public VisualElement(Lookup lkp) {
            obj = lkp.lookup(InvalidDataObject.class);
            assert obj != null;
            initComponents();
        }

        private void initComponents() {
            JPanel contentPanel = new JPanel(new MigLayout("wrap 1"));
            Detail detail = obj.getDetail();
            contentPanel.add(new JLabel(detail.getMessage()));
            if (detail.getThrowable() != null) {
                contentPanel.add(new JLabel(detail.getThrowable().toString()));
            }
            // avoid showing a stacktrace to the user for now, it will be in the logfile.
            
            bodyPanel = BodyPanel.create(contentPanel);
            bodyPanel.setTitle(obj.getPrimaryFile().getNameExt());
            bodyPanel.setIcon(ImageUtilities.loadImageIcon(ERROR_ICON, true));
        }

        @Override
        public JComponent getVisualRepresentation() {
            return bodyPanel;
        }

        @Override
        public JComponent getToolbarRepresentation() {
            return toolbar;
        }

        @Override
        public Action[] getActions() {
            return new Action[0];
        }

        @Override
        public Lookup getLookup() {
            return obj.getLookup();
        }

        @Override
        public void componentOpened() {
        }

        @Override
        public void componentClosed() {
        }

        @Override
        public void componentShowing() {
            if (callback != null) {
                updateName();
            }
        }

        public void updateName() {
            if (callback != null) {
                String name = obj.getName();
                TopComponent tc = callback.getTopComponent();
                tc.setHtmlDisplayName(name);
                tc.setDisplayName(name);
                tc.setName(name);
                tc.setToolTipText("");
            }
        }
        
        @Override
        public void componentHidden() {
        }

        @Override
        public void componentActivated() {
        }

        @Override
        public void componentDeactivated() {
        }

        @Override
        public UndoRedo getUndoRedo() {
            return UndoRedo.NONE;
        }

        @Override
        public void setMultiViewCallback(MultiViewElementCallback callback) {
            this.callback = callback;
        }

        @Override
        public CloseOperationState canCloseElement() {
            return CloseOperationState.STATE_OK;
        }
    }
}
