/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import com.vividsolutions.jts.geom.Coordinate;
import eu.agrosense.client.grid.util.AggregationGrid;
import eu.agrosense.client.grid.util.RasterHelper;
import java.awt.Point;
import java.awt.image.DataBuffer;
import java.awt.image.WritableRaster;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import nl.cloudfarming.client.geoviewer.Grid;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.DeviceElement;
import nl.cloudfarming.client.isobus.model.IsobusGrid;
import nl.cloudfarming.client.isobus.model.IsobusTime;
import nl.cloudfarming.client.isobus.model.util.TimeLogUtil;
import nl.cloudfarming.client.isobus.type.TimeLogGridTransformationStrategies.TimeLogTransformationContext;
import nl.cloudfarming.client.util.Callback;
import org.geotools.coverage.CoverageFactoryFinder;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridCoverageFactory;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.geometry.BoundingBox;
/**
 * Raster factory for Isobus grids.
 * 
 * Both GRD (types 1 and 2, via intermediate value grid) and TLG (for rate and yield DDIs) binary files are supported.
 * 
 * @author Merijn Zengers
 */
public class RasterFactory {
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.type.RasterFactory");

    /**
     * Creates a raster from a grid of raw process data values.
     * 
     * @param grid the grid to create a raster for (matches the grid tag in taskdata.xml)
     * @param valueGrid data grid of raw ProcessDataVariable values;
     * 
     * @return the raster created
     */
    public static GridCoverage2D createRaster(IsobusGrid grid, IsobusValueGrid valueGrid, Callback<double[]> rangeCallback) {
        assert grid.getMaximumColumn() * grid.getMaximumRow() == valueGrid.getSize();
        
        double minx = grid.getMinimumEastPosition().doubleValue();
        double maxx = grid.getMinimumEastPosition().doubleValue() + (grid.getCellEastSize() * grid.getMaximumColumn());
        double miny = grid.getMinimumNorthPosition().doubleValue() ;
        double maxy = grid.getMinimumNorthPosition().doubleValue() + (grid.getCellNorthSize() * grid.getMaximumRow());

        int rows = valueGrid.getNumberOfRows();
        int columns = valueGrid.getNumberOfColumns();
        int bands = 1;
        Point location = new Point(0, 0);
        WritableRaster raster = WritableRaster.createBandedRaster(DataBuffer.TYPE_INT, columns, rows, bands, location);

        raster.setSamples(0, 0, columns, rows, 0, valueGrid.getValues());

        BoundingBox boundingBox = new ReferencedEnvelope(minx, maxx, miny, maxy, DefaultGeographicCRS.WGS84);

        handleRangeCallback(raster, rangeCallback);
        
        GridCoverageFactory factory = CoverageFactoryFinder.getGridCoverageFactory(null);
        return factory.create("name", raster, boundingBox);
    }
    
    /**
     * Creates a raster from a timelog binary/xml file pair.
     *
     * @param binary timelog binary file.
     * @param ddiToPlot 
     * @param rangeCallback optional callback, when not null, and if a raster
     * is actually created, the callback will be called with the minimum and 
     * maximum value present in the grid samples.
     * @return 
     */
    public static GridCoverage2D createRaster(TimeLogBinaryDataObject binary, DDI ddiToPlot, Callback<double[]> rangeCallback) {
        LOGGER.log(Level.INFO, "Creating grid coverage for file {0}", binary.getName());
        final List<IsobusTime> isobusTimes = binary.getIsobusTimes();
        final List<GpsDerivedValue> gpsDerivedValues = binary.getGpsDerivedValues();
        final TimeLogXMLDataObject xml = binary.getXmlHeader();
        
        if (ddiToPlot == null) {
            LOGGER.warning("cannot create raster: no DDI to show");
            return null;
        }
        if (isobusTimes == null || isobusTimes.isEmpty()) {
            LOGGER.warning("cannot create raster: no binary timelog records available");
            return null;
        }
        if (gpsDerivedValues == null || gpsDerivedValues.isEmpty()) {
            LOGGER.warning("cannot create raster: no GPS derived values available");
            return null;            
        }
        if (xml.getDataLogValues() == null || xml.getDataLogValues().isEmpty()) {
            // binary contains only the time and position measurements.
            LOGGER.warning("cannot create raster: no DataLogValue metadata available in xml header");
            return null;            
        }

        RasterHelper rasterHelper = new TimeLogRasterHelper(isobusTimes, getMaxWorkwidth(binary));
        if (!rasterHelper.hasValidEnvelope()) {
            LOGGER.warning("cannot create raster: not enough valid GPS positions found");
            return null;
        }
        
        AggregationGrid<TimeLogTransformationContext> aggregationGrid = TimeLogGridTransformationStrategies.getAggregationGrid(ddiToPlot, binary, rasterHelper);
        TimeLogGridCalculationStrategy calculationStrategy = TimeLogGridCalculationStrategies.getStrategy(binary, aggregationGrid, ddiToPlot, rasterHelper);
        
        final int len = isobusTimes.size();
        
        for (int i = 0; i < len; i++) {
            // record preconditions:
            IsobusTime isobusTime = isobusTimes.get(i);
            if (!Positions.isUsable(isobusTime.getPTN())) { continue; }
            GpsDerivedValue gpsDerivedValue = gpsDerivedValues.get(i); // bearing required for section positions, distance for affected areas.
            if (gpsDerivedValue == null) { continue; }

            // process one record:
            calculationStrategy.handle(isobusTime, gpsDerivedValue);
        }
        
        // fill the raster with the aggregated results and create the coverage:
        rasterHelper.setSamples(aggregationGrid.getIntValues());
        handleRangeCallback(rasterHelper.getRaster(), rangeCallback);
        return rasterHelper.createCoverage();
    }
    
    // currently used to inform legend palettes of the min/max values in the grid
    private static void handleRangeCallback(WritableRaster raster, Callback<double[]> rangeCallback) {
        if (rangeCallback != null) {
            int w = raster.getWidth(), h = raster.getHeight();
            int[] buf = raster.getSamples(0, 0, w, h, 0, new int[w*h]);
            Integer min = null, max = null;
            for (int value : buf) {
                if (value != Grid.EMPTY_VALUE) {
                    if (min == null) {
                        min = max = value;
                    } else if (value < min) {
                        min = value;
                    } else if (value > max) {
                        max = value;
                    }
                }
            }
            if (min != null) {
                rangeCallback.call(new double[] {min.doubleValue(), max.doubleValue()});
            }
        }
    }
    
    private static double getMaxWorkwidth(TimeLogBinaryDataObject binary) {
        DeviceElement deviceElement = TimeLogUtil.getMainDeviceElement(binary.getXmlHeader().getDeviceElements());
        Double max = null;
        // use values from binary if available:
        if (binary.getXmlHeader().getDDIs(deviceElement).contains(DDI.DDI_67)) {
            max = TimeLogUtil.getMaxProcessDataValueForDeviceElement(binary.getIsobusTimes(), deviceElement, DDI.DDI_67);
        }
        // else check default property:
        if (max == null) {
            max = TimeLogUtil.getDevicePropertyValue(binary.getTaskData(), deviceElement, DDI.DDI_68);
        }
        return max == null ? 0 : max;
    }
    
    static class TimeLogRasterHelper extends RasterHelper {
        final List<IsobusTime> isobusTimes;
        final double maxDeviceWorkWidth; // in mm
        
        public TimeLogRasterHelper(List<IsobusTime> isobusTimes, double maxDeviceWorkWidth) {
            this.isobusTimes = isobusTimes;
            this.maxDeviceWorkWidth = maxDeviceWorkWidth;
            initEnvelopeCoordinates();
            
            if (hasValidEnvelope()) {
                applyWorkWidth();
                initDimensions(isobusTimes.size());
            }
        }
        
        @Override
        protected final void initEnvelopeCoordinates() {
            // first loop until a valid is PTN found, then loop rest to compare
            for (IsobusTime isobusTime : isobusTimes) {
                if (!Positions.isUsable(isobusTime.getPTN())) { continue; }
                
                double lon = isobusTime.getPTN().getPositionEast().doubleValue();
                double lat = isobusTime.getPTN().getPositionNorth().doubleValue();
                if (minx == null) { // first valid PTN, initialize:
                    minx = lon;
                    miny = lat;
                    maxx = lon;
                    maxy = lat;
                } else { // other valid PTN's, compare:
                    minx = (minx < lon) ? minx : lon;
                    miny = (miny < lat) ? miny : lat;
                    maxx = (maxx > lon) ? maxx : lon;
                    maxy = (maxy > lat) ? maxy : lat;                
                }
            }
        }

        // enlarge bounding box to account for device working width;
        // apply full width on both sides to be safe.
        private void applyWorkWidth() {
            if (maxDeviceWorkWidth > 0.0) {
                // coord (x, y) = (lon, lat)
                // offset (x, y) relative to driving direction = (forward, right hand side)
                minx = TimeLogUtil.getCoordinate(new Coordinate(minx, miny), 0, -maxDeviceWorkWidth, 0).x;
                maxx = TimeLogUtil.getCoordinate(new Coordinate(maxx, maxy), 0, maxDeviceWorkWidth, 0).x;                
                miny = TimeLogUtil.getCoordinate(new Coordinate(minx, miny), 0, maxDeviceWorkWidth, 90).y;
                maxy = TimeLogUtil.getCoordinate(new Coordinate(maxx, maxy), 0, -maxDeviceWorkWidth, 90).y;                
            }
        }
    }

}
