/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import nl.cloudfarming.client.isobus.model.*;
import org.openide.awt.NotificationDisplayer;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.FileEntry;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiDataObject.Entry;
import org.openide.loaders.MultiFileLoader;
import org.openide.util.ImageUtilities;
import org.openide.util.NbBundle;

/**
 * File loader for TASKDATA files TASKDATA files are in the format TASKDATA.XML
 * or this format in lower case
 *
 * Mandatory files : TASKDATA.XML or this format in lower case —the task data
 * file
 *
 * Optional files : TLG[0-9][0-9][0-9][0-9][0-9].BIN or this format in lower
 * case — contains time log values; TLG[0-9][0-9][0-9][0-9][0-9].XML or this
 * format in lower case — contains time log header;
 * GRD[0-9][0-9][0-9][0-9][0-9].BIN or this format in lower case — grid format;
 *
 * @author Merijn Zengers
 */
@NbBundle.Messages({
    "# {0} - file",
    "TaskDataFileLoader file not loaded title=File {0} was not loaded,",
})
public class TaskDataFileLoader extends MultiFileLoader {
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.type.TaskDataFileLoader");
    private static final String ERROR_ICON = "nl/cloudfarming/client/util/swing/error.png";

    public static final String PRIMARY_PATTERN = "(?i)taskdata\\.xml";
    // secondary files:
    private static final String GRID_PATTERN = GridFileLoader.PRIMARY_PATTERN;
    private static final String TIMELOG_PATTERN_BIN = TimeLogBinaryFileLoader.PRIMARY_PATTERN;
    private static final String TIMELOG_PATTERN_XML = TimeLogXMLFileLoader.PRIMARY_PATTERN;
    
    public TaskDataFileLoader() {
        super(TaskDataObject.class.getName());
    }

    @Override
    protected String actionsContext() {
        StringBuilder builder = new StringBuilder("Loaders/");
        builder.append(TaskDataObject.MIME_TYPE);
        builder.append("/Actions/");
        return builder.toString();
    }

    /**
     * Finds the primary file by one of the two secondary files The secondary
     * files are GRD[0-9][0-9][0-9][0-9][0-9].BIN or toLower And
     * TLG[0-9][0-9][0-9][0-9][0-9].BIN or toLower
     *
     * @param fo the fileObject to find the primary file by
     * @return if fo is one of the secondary files returns the primary file of
     * format TASKDATA.XML or toLower
     */
    @Override
    protected FileObject findPrimaryFile(FileObject fo) {
        // never recognize folders.
        if (fo.isFolder()) {
            return null;
        }
        Pattern primaryPattern = Pattern.compile(PRIMARY_PATTERN);
        Pattern gridPattern = Pattern.compile(GRID_PATTERN);
        Pattern timeLogPatternBin = Pattern.compile(TIMELOG_PATTERN_BIN);
        Pattern timeLogPatternXml = Pattern.compile(TIMELOG_PATTERN_XML);
        if (primaryPattern.matcher(fo.getNameExt()).find()) {
            return fo;
        } else if (gridPattern.matcher(fo.getNameExt()).find() || timeLogPatternBin.matcher(fo.getNameExt()).find() || timeLogPatternXml.matcher(fo.getNameExt()).find()) {
            for (FileObject child : fo.getParent().getChildren()) {
                if (!child.equals(fo)) {
                    Pattern childPattern = Pattern.compile(PRIMARY_PATTERN);
                    if (childPattern.matcher(child.getNameExt()).find()) {
                        return child;
                    }
                }
            }
        }
        LOGGER.log(Level.FINEST, "No primary file found for FileObject {0}", fo);
        return null;
    }

    /**
     * Creates the MultidataObject for the file with format TASKDATA.XML or
     * toLower() Will add GRID and TimeLog files as secondary files if present
     * Only ads secondary files if they are referenced in the primary file
     *
     * @param primaryFile the Primary file to create the DataObject for
     * @return the created DataObject
     * @throws DataObjectExistsException if the DataObject already exists
     * @throws IOException
     */
    @Override
    protected MultiDataObject createMultiObject(FileObject primaryFile) throws DataObjectExistsException, IOException {
        List<FileObject> secondary = new ArrayList<>();
        IsobusTaskData isobusTaskData;
        
        try {
            isobusTaskData = new IsobusMarshaller().unMarshall(primaryFile);
        } catch (IsobusUnmarshallingException ex) {
            // include path, as filename will always be taskdata.xml
            String message = Bundle.TaskDataFileLoader_file_not_loaded_title(primaryFile.getPath());
            NotificationDisplayer.getDefault().notify(message, ImageUtilities.loadImageIcon(ERROR_ICON, true), ex.toString(), null);
            InvalidDataObject.Detail invalidDataDetail = InvalidDataObject.detail(message, ex);
            return new InvalidDataObject(primaryFile, this, TaskDataObject.ICON_BASE, invalidDataDetail);
        }
        
        for (FileObject child : primaryFile.getParent().getChildren()) {
            if (!child.equals(primaryFile)) {
                //Check every task in the taskdata for grid and timelog file references:
                for (IsobusTask task : isobusTaskData.getTSK()) {
                    if (task.getGRD() != null && task.getGRD().getFilename().equalsIgnoreCase(child.getName())) {
                        secondary.add(child);
                    }
                    for (TimeLog timeLog : task.getTLG()) {
                        if (timeLog.getFilename().equalsIgnoreCase(child.getName())) {
                            secondary.add(child);
                        }
                    }
                }
            }
        }

        return new TaskDataObject(primaryFile, this, isobusTaskData, secondary);
    }

    /**
     * Creates a primary entry for the primary files
     *
     * @param obj the dataObject to create a primary entry for
     * @param primaryFile the primary file to add as primary entry
     * @return the created primary entry
     */
    @Override
    protected Entry createPrimaryEntry(MultiDataObject obj, FileObject primaryFile) {
        Entry entry = new FileEntry(obj, primaryFile);
        LOGGER.log(Level.FINEST, "Creating primary entry for {0}", primaryFile.getName());
        return entry;
    }

    /**
     * Creates a secondary entry for the provided MultiDataObject
     *
     * @param obj the MultiDataObject to create the secondary entry for
     * @param secondaryFile the secondary file to create the entry from
     * @return The created entry
     */
    @Override
    protected Entry createSecondaryEntry(MultiDataObject obj, FileObject secondaryFile) {
        Pattern gridPattern = Pattern.compile(GRID_PATTERN);
        Pattern timeLogPatternBin = Pattern.compile(TIMELOG_PATTERN_BIN);
        Pattern timeLogPatternXml = Pattern.compile(TIMELOG_PATTERN_XML);
        assert (gridPattern.matcher(secondaryFile.getNameExt()).find() || timeLogPatternBin.matcher(secondaryFile.getNameExt()).find() || timeLogPatternXml.matcher(secondaryFile.getNameExt()).find());

        FileEntry entry = new FileEntry(obj, secondaryFile);
        LOGGER.log(Level.FINEST, "Creating secondary entry for {0}", secondaryFile.getName());
        return entry;
    }
}
