/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;
import nl.cloudfarming.client.isobus.model.*;
import nl.cloudfarming.client.isobus.tree.TaskDataRootNode;
import org.netbeans.core.spi.multiview.MultiViewElement;
import org.netbeans.core.spi.multiview.text.MultiViewEditorElement;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiFileLoader;
import org.openide.nodes.Node;
import org.openide.util.HelpCtx;
import org.openide.util.Lookup;
import org.openide.util.NbBundle.Messages;
import org.openide.windows.TopComponent;

/**
 * DataObject representing isobus taskdata.
 *
 * <p>The TaskDataObject is a MultiDataObject, the TASKDATA.XML contain
 * references to binary files for grids and timelogs. Those linked binary files
 * are contained as secondary files in this dataobject. In order to be able to
 * link the secondary files, the taskdata,xml file needs to be parsed, which can
 * be slow.</p>
 *
 * @author Timon Veenstra <monezz@gmail.com>
 */
public class TaskDataObject extends MultiDataObject {

    public static final String MIME_TYPE = "application/x-taskdata";
    public static final String ICON_BASE = "nl/cloudfarming/client/isobus/type/iso.png";
    
    private IsobusTaskData isobusTaskData;
    /*
     * Workaround:
     * 
     * All loading logic has been moved into the FileLoaders.
     * Processing of GRD files needs access to parts of the taskdata.xml via
     * {@code getGrid(FileObject)} and {@code getTreatmentzones(FileObject)}.
     * Both methods used to check whether the passed file is a secondary entry 
     * of the taskfile via {@code MultiDataObject.findSecondaryEntry(FileObject)}.
     * This will however also call DataObject.find(FileObject) internally 
     * (via {@code MultiFileLoader.checkFiles()}), on the GRD file that's currently 
     * being loaded, causing it to call getGrid() and getTreatmentzones() again,
     * and eventually causing a StackOverflowError.
     * Long story short: we keep our own list of secondary files so we can check
     * without turning them into DataObjects yet.
     */
    private final List<FileObject> secondaryFileObjects = new ArrayList<>();

    public TaskDataObject(FileObject pf, MultiFileLoader loader, IsobusTaskData isobusTaskData, List<FileObject> secondary) throws DataObjectExistsException, IOException {
        super(pf, loader);
        registerEditor(MIME_TYPE, true);
        this.isobusTaskData = isobusTaskData;

        setSecondaryFiles(secondary);
    }

    /**
     * Method to set the secondary files
     *
     * @param secondary the secondary files
     */
    private void setSecondaryFiles(List<FileObject> secondary) {
        this.secondaryFileObjects.addAll(secondary);
        for (FileObject fileObject : secondary) {
            registerEntry(fileObject);
        }
    }

    /**
     * Get the IsobusTaskData from this DataObject
     *
     * @return IsobusTaskData object for this DataObject
     */
    public IsobusTaskData getIsobusTaskData() {
        return this.isobusTaskData;
    }

    /**
     * Gets the isobus grid object from the taskdata matching the supplied grid
     * binary file Will throw AssertionExceptions if the supplied file does not
     * match the grid filename format of GRD[0-9][0-9][0-9][0-9][0-9].BIN or if
     * the supplied file is not a secondary file of this TaskDataObject
     *
     * @param gridFile the grid binary file to search the isobus grid for
     * @return The IsobusTreatmentZoneGrid object if found else null
     */
    public IsobusGrid getGrid(FileObject gridFile) {
//        assert findSecondaryEntry(gridFile) != null : "The supplied file is not secondary file of this TaskDataDataObject";
        assert secondaryFileObjects.contains(gridFile) : "The supplied file is not secondary file of this TaskDataDataObject";
        Pattern gridPattern = Pattern.compile(GridFileLoader.PRIMARY_PATTERN);
        assert gridPattern.matcher(gridFile.getNameExt()).find() : "The file name of the Suplied file does not match the grid filename format of GRD[0-9][0-9][0-9][0-9][0-9].BIN";
        IsobusGrid grid = null;
        for (IsobusTask task : isobusTaskData.getTSK()) {
            if (task.getGRD() != null && task.getGRD().getFilename() != null && task.getGRD().getFilename().toLowerCase().equals(gridFile.getName().toLowerCase())) {
                grid = task.getGRD();
            }
        }
        return grid;
    }

    /**
     * Gets the isobus treatment zones referenced in the supplied grid file Will
     * throw AssertionExceptions if the supplied file does not match the grid
     * filename format of GRD[0-9][0-9][0-9][0-9][0-9].BIN or if the supplied
     * file is not a secondary file of this TaskDataObject
     *
     * @param gridFile the grid file to search the treatment zones for
     * @return a list of found treatment zones or null if nothing is found
     */
    public List<IsobusTreatmentZone> getTreatmentzones(FileObject gridFile) {
        assert secondaryFileObjects.contains(gridFile) : "The supplied file is not secondary file of this TaskDataDataObject";
//        assert findSecondaryEntry(gridFile) != null : "The supplied file is not secondary file of this TaskDataDataObject";
        Pattern gridPattern = Pattern.compile(GridFileLoader.PRIMARY_PATTERN);
        assert gridPattern.matcher(gridFile.getNameExt()).find() : "The file name of the Suplied file does not match the grid filename format of GRD[0-9][0-9][0-9][0-9][0-9].BIN";
        List<IsobusTreatmentZone> treatmentZones = null;
        for (IsobusTask task : isobusTaskData.getTSK()) {
            if (task.getGRD() != null && task.getGRD().getFilename() != null && task.getGRD().getFilename().toLowerCase().equals(gridFile.getName().toLowerCase())) {
                treatmentZones = task.getTZN();
            }
        }
        return treatmentZones;
    }

    @Override
    protected Node createNodeDelegate() {
        //FIXME quick workaround for AGROSENSE-582, investigate and apply proper fix
        if (isobusTaskData != null) {
            return new TaskDataRootNode(isobusTaskData, super.createNodeDelegate());
        } else {
            return super.createNodeDelegate();
        }
    }

    @Override
    public HelpCtx getHelpCtx() {
        return new HelpCtx("nl.cloudfarming.client.isobus.project.explorer");
    }

    @Override
    protected int associateLookup() {
        return 1;
    }

    @MultiViewElement.Registration(displayName = "#LBL_Isobus_EDITOR",
    iconBase = "nl/cloudfarming/client/isobus/type/iso.png",
    mimeType = "application/x-taskdata",
    persistenceType = TopComponent.PERSISTENCE_ONLY_OPENED,
    preferredID = "Isobus",
    position = 1000)
    @Messages("LBL_Isobus_EDITOR=Source")
    public static MultiViewEditorElement createEditor(Lookup lkp) {
        return new MultiViewEditorElement(lkp);
    }
}
