/*
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *    johan - initial API and implementation and/or initial documentation
 */
package nl.cloudfarming.client.isobus.type;

import com.vividsolutions.jts.geom.Coordinate;
import eu.agrosense.client.grid.util.AggregationGrid;
import eu.agrosense.client.grid.util.RasterHelper;
import java.util.Arrays;
import java.util.Map;
import java.util.Set;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.DeviceElement;
import nl.cloudfarming.client.isobus.model.IsobusTime;
import nl.cloudfarming.client.isobus.model.util.TimeLogUtil;

/**
 *
 * @author johan
 */
public class TimeLogGridCalculationStrategies {

    private TimeLogGridCalculationStrategies() {}

    /**
     * Find an appropriate calculation strategy for a specific DDI.
     * 
     * @param binary
     * @param aggregationGrid
     * @param ddi
     * @param rasterHelper
     * @return 
     */
    public static TimeLogGridCalculationStrategy getStrategy(TimeLogBinaryDataObject binary, 
            AggregationGrid<TimeLogGridTransformationStrategies.TimeLogTransformationContext> aggregationGrid, DDI ddi, RasterHelper rasterHelper) {
        
        if (Arrays.asList(TimeLogUtil.ACTUAL_RATE_DDIS).contains(ddi)
                || Arrays.asList(TimeLogUtil.YIELD_DDIS).contains(ddi)) {
            return new AppliedRateYieldStrategy(binary, aggregationGrid, ddi, rasterHelper);
        }
        if (Arrays.asList(TimeLogUtil.SETPOINT_RATE_DDIS).contains(ddi)) {
            return new SetpointRateStrategy(binary, aggregationGrid, ddi, rasterHelper);
        }
        return new NullStrategy(binary, aggregationGrid, ddi, rasterHelper);
    }
    
    // fallback which won't produce any values.
    static final class NullStrategy extends TimeLogGridCalculationStrategy {

        public NullStrategy(TimeLogBinaryDataObject binary, AggregationGrid<TimeLogGridTransformationStrategies.TimeLogTransformationContext> aggStrategy, DDI ddi, RasterHelper rasterHelper) {
            super(binary, aggStrategy, ddi, rasterHelper);
        }

        @Override public void handle(IsobusTime isobusTime, GpsDerivedValue gpsDerivedValue) {}
        @Override protected void handleInternal(IsobusTime isobusTime, GpsDerivedValue gpsDerivedValue) {}
    }
    
    static class SetpointRateStrategy extends TimeLogGridCalculationStrategy {
        final Set<DeviceElement> sections;
        final DeviceElement mainDevice;
        
        public SetpointRateStrategy(TimeLogBinaryDataObject binary, AggregationGrid<TimeLogGridTransformationStrategies.TimeLogTransformationContext> aggStrategy, DDI ddi, RasterHelper rasterHelper) {
            super(binary, aggStrategy, ddi, rasterHelper);
            Set<DeviceElement> deviceElements = binary.getXmlHeader().getDeviceElements();
            mainDevice = TimeLogUtil.getMainDeviceElement(deviceElements);
            sections = TimeLogUtil.getSectionDeviceElements(deviceElements);
        }

        @Override
        protected void handleInternal(IsobusTime isobusTime, GpsDerivedValue gpsDerivedValue) {
            Double rate = TimeLogUtil.getProcessDataValueForDeviceElement(isobusTime.getDLV(), mainDevice, ddiToPlot);
            // apply rate to all sections (when available) to fill up the grid, values will get averaged over cell.
            if (rate != null && rate != 0.0) {
                if (sections.isEmpty()) {
                    add(origin, rate, mainDevice);
                } else {
                    for (DeviceElement section :sections) {
                        double xOffset = TimeLogUtil.getXOffsetForDeviceElement(isobusTime, section);
                        double yOffset = TimeLogUtil.getYOffsetForDeviceElement(isobusTime, section);
                        Coordinate sectionCoordinate = TimeLogUtil.getCoordinate(origin, xOffset, yOffset, gpsDerivedValue.getBearing());
                        add(sectionCoordinate, rate, section);
                    }
                }
            }
        }
    }
    
    static class AppliedRateYieldStrategy extends TimeLogGridCalculationStrategy {
        final boolean hasSections;
        final boolean useCondensedWorkState;
        final Set<DeviceElement> deviceElements;
        final DeviceElement mainDevice;
        final boolean distributeAmounts;
        
        public AppliedRateYieldStrategy(TimeLogBinaryDataObject binary, AggregationGrid<TimeLogGridTransformationStrategies.TimeLogTransformationContext> aggStrategy, DDI ddi, RasterHelper rasterHelper) {
            super(binary, aggStrategy, ddi, rasterHelper);
            TimeLogXMLDataObject xml = binary.getXmlHeader();
            deviceElements = xml.getDeviceElements();
            mainDevice = TimeLogUtil.getMainDeviceElement(deviceElements);
            hasSections = !TimeLogUtil.getSectionDeviceElements(deviceElements).isEmpty();
            useCondensedWorkState = xml.getDDIs().contains(DDI.DDI_161);
            distributeAmounts = false; // TODO: depends on UOM type (absolute / per area / ...), determine from same DDI subsets that control the aggregation strategy
        }

        @Override
        protected void handleInternal(IsobusTime isobusTime, GpsDerivedValue gpsDerivedValue) {
            if (!hasSections) {
                // v1) simple version plotting the total application amount at the vehicle origin:

                // check if main device is on:
                Double workState = TimeLogUtil.getActualWorkStateForDeviceElement(isobusTime.getDLV(), mainDevice);
                if (workState != TimeLogUtil.WORK_STATE_ON_AUTO) { return; }
             
                Double rate = TimeLogUtil.getProcessDataValueForDeviceElement(isobusTime.getDLV(), mainDevice, ddiToPlot);
                if (rate != null && rate != 0.0) {
                    // TODO: get main device location.
                    // the offsets can also be registered as device properties
                    // in which case they're constant for the entire file (i.e. set a flag in the contructor)
                    add(origin, rate, mainDevice);
                }
            } else {
                // v2) split application amount over active sections, use section offsets:
                Map<DeviceElement, Double> rateMap = useCondensedWorkState ?
                        TimeLogUtil.getCondensedWorkStateApplicationRate(deviceElements, isobusTime.getDLV(), ddiToPlot, distributeAmounts) :
                        TimeLogUtil.getVerboseWorkStateApplicationRate(deviceElements, isobusTime.getDLV(), ddiToPlot, distributeAmounts);

                if (rateMap != null && !rateMap.isEmpty()) {
                    for (Map.Entry<DeviceElement, Double> entry : rateMap.entrySet()) {
                        Double rate = entry.getValue();
                        if (rate != null && rate != 0.0) {
                            DeviceElement section = entry.getKey();
                            double xOffset = TimeLogUtil.getXOffsetForDeviceElement(isobusTime, section);
                            double yOffset = TimeLogUtil.getYOffsetForDeviceElement(isobusTime, section);
                            Coordinate sectionCoordinate = TimeLogUtil.getCoordinate(origin, xOffset, yOffset, gpsDerivedValue.getBearing());
                            add(sectionCoordinate, rate, section);
                        }
                    }
                }
            } 
        }
        
    }
}
