/*
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *    johan - initial API and implementation and/or initial documentation
 */
package nl.cloudfarming.client.isobus.type;

import eu.agrosense.client.grid.util.AggregationGrid;
import eu.agrosense.client.grid.util.AggregationStrategies;
import eu.agrosense.client.grid.util.AggregationStrategy;
import eu.agrosense.client.grid.util.RasterHelper;
import eu.agrosense.client.grid.util.TransformationContext;
import eu.agrosense.client.grid.util.TransformationStrategies;
import eu.agrosense.client.grid.util.TransformationStrategy;
import java.util.Arrays;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.DeviceElement;
import nl.cloudfarming.client.isobus.model.DeviceElementTypeEnum;
import nl.cloudfarming.client.isobus.model.IsobusTaskData;
import nl.cloudfarming.client.isobus.model.IsobusTime;
import nl.cloudfarming.client.isobus.model.util.TimeLogUtil;

/**
 * Transformation strategies for different types of measurements found in timelog files.
 * 
 * <p>
 * Including factory methods to create an aggregation grid and pick the 
 * appropriate transformation and aggregation strategies based on DDI.</p>
 * 
 * @author johan
 */
public class TimeLogGridTransformationStrategies {
    public static final DDI WORK_WIDTH_DDI = DDI.DDI_67;
    public static final DDI DEFAULT_WORK_WIDTH_DDI = DDI.DDI_68;
    // subset of the TimeLogUtil ACTUAL_RATE_DDIS and YIELD_DDIS which have a "per area" UOM:
    public static final EnumSet<DDI> ACTUAL_RATE_YIELD_PER_AREA_DDIS = EnumSet.of(DDI.DDI_2, DDI.DDI_7, DDI.DDI_12, DDI.DDI_83, DDI.DDI_84, DDI.DDI_85, DDI.DDI_181);

    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.isobus.type.TimeLogGridTransformationStrategies");
    
    private TimeLogGridTransformationStrategies() {}
    
    /**
     * POJO context used while processing a single value into a cell aggregate.
     */
    public static class TimeLogTransformationContext implements TransformationContext {
        private DeviceElement deviceElement;
        private IsobusTime isobusTime;
        private GpsDerivedValue gpsDerivedValue;

        public DeviceElement getDeviceElement() {
            return deviceElement;
        }

        // main device or section
        public void setDeviceElement(DeviceElement deviceElement) {
            this.deviceElement = deviceElement;
        }

        public IsobusTime getIsobusTime() {
            return isobusTime;
        }

        public void setIsobusTime(IsobusTime isobusTime) {
            this.isobusTime = isobusTime;
        }

        public GpsDerivedValue getGpsDerivedValue() {
            return gpsDerivedValue;
        }

        public void setGpsDerivedValue(GpsDerivedValue gpsDerivedValue) {
            this.gpsDerivedValue = gpsDerivedValue;
        }
    }
    
    public static abstract class TimeLogTransformationStrategy extends TransformationStrategies.WithRasterContext<TimeLogTransformationContext> {

        public TimeLogTransformationStrategy(DDI ddi, RasterHelper rasterHelper) {
            super(rasterHelper);
        }
    }
    
    /**
     * Find an appropriate transformation strategy for a specific DDI.
     * 
     * @param ddi
     * @param binary 
     * @param rasterHelper
     * @return 
     */
    protected static TransformationStrategy<TimeLogTransformationContext> getTransformationStrategy(DDI ddi, TimeLogBinaryDataObject binary, RasterHelper rasterHelper) {
        
        // actual/applied rates and yields, per area:
        if (ACTUAL_RATE_YIELD_PER_AREA_DDIS.contains(ddi)) {
            return new AppliedRatePerAreaStrategy(ddi, rasterHelper, binary);
        }

        // all setpoint rates: use as is.
        if (Arrays.asList(TimeLogUtil.SETPOINT_RATE_DDIS).contains(ddi)) {
            return new TransformationStrategies.Noop<>();
        }        

        return new TransformationStrategies.Noop<>();
    }
    
    /**
     * Find an appropriate aggregation strategy for a specific DDI.
     * 
     * @param ddi
     * @return a factory that the grid will use to create the strategy instances
     */
    protected static AggregationStrategy.Factory getAggregationStrategyFactory(DDI ddi) {
        if (ACTUAL_RATE_YIELD_PER_AREA_DDIS.contains(ddi)) {
            return AggregationStrategies.sum();
        }

        if (Arrays.asList(TimeLogUtil.SETPOINT_RATE_DDIS).contains(ddi)) {
            return AggregationStrategies.average();
        }        

        return AggregationStrategies.sum();
    }
    
    /**
     * Create an aggregation grid for a specific DDI, including the appropriate
     * transformation and aggregation strategies.
     * 
     * @param ddi
     * @param binary 
     * @param rasterHelper
     * @return 
     */
    public static AggregationGrid<TimeLogTransformationContext> getAggregationGrid(DDI ddi, TimeLogBinaryDataObject binary, RasterHelper rasterHelper) {

        return new AggregationGrid<>(
                rasterHelper.getNumberOfColumns(), rasterHelper.getNumberOfRows(),
                getTransformationStrategy(ddi, binary, rasterHelper), 
                getAggregationStrategyFactory(ddi));
    }
    
    // corresponds to a DOR pointing to either a DPD or a DPT:
    private enum WorkWidthSource { PROCESS_DATA, PROPERTY; }
    
    /**
     * Transformation strategy for applied rates and yields using a "per area" unit of measure.
     * 
     * Get (estimate) the area a rate was applied to from the working width and
     * the distance driven for a single timelog record (from interpolated GPS values).
     * Calculate the applied amount from the affected area and the rate.
     * These amounts will be summed in the intermediate aggregation grid.
     * The rate used in the grid coverage will be the total amount over the gridcell area.
     */
    private static class AppliedRatePerAreaStrategy extends TimeLogTransformationStrategy {
        private final Map<DeviceElement,WorkWidthSource> wwSources = new HashMap<>();
        private final Map<DeviceElement,Double> wwPropertyValues = new HashMap<>();
        
        public AppliedRatePerAreaStrategy(DDI ddi, RasterHelper rasterHelper, TimeLogBinaryDataObject binary) {
            super(ddi, rasterHelper);
            // determine where to find the workwidth for each device element:
            initWorkWidthSources(binary);
        }
        
        private void initWorkWidthSources(TimeLogBinaryDataObject binary) {
            EnumSet<DeviceElementTypeEnum> types = EnumSet.of(DeviceElementTypeEnum.DEVICE, DeviceElementTypeEnum.SECTION);
            TimeLogXMLDataObject xml = binary.getXmlHeader();
            IsobusTaskData taskdata = binary.getTaskData();
            Double value;
            for (DeviceElement det : xml.getDeviceElements()) {
                if (types.contains(det.getType())) {
                    if (xml.getDDIs(det).contains(WORK_WIDTH_DDI)) {
                        wwSources.put(det, WorkWidthSource.PROCESS_DATA);
                        continue;
                    }
                    value = TimeLogUtil.getDevicePropertyValue(taskdata, det, WORK_WIDTH_DDI);
                    if (value != null) {
                        wwSources.put(det, WorkWidthSource.PROPERTY);
                        wwPropertyValues.put(det, value);
                        continue;
                    }
                    value = TimeLogUtil.getDevicePropertyValue(taskdata, det, DEFAULT_WORK_WIDTH_DDI);
                    if (value != null) {
                        wwSources.put(det, WorkWidthSource.PROPERTY);
                        wwPropertyValues.put(det, value);
                        continue;
                    }
                    LOGGER.log(Level.WARNING, "Could not determine a workwidth source for device element {0}", det.getId());
                }
            }
        }

        @Override
        public Double transformIn(double value, TimeLogTransformationContext context) {
            DeviceElement deviceElement = context.getDeviceElement();
            // workwidth in mm:
            Double workWidth = null;
            switch (wwSources.get(deviceElement)) {
                case PROCESS_DATA:
                    workWidth = TimeLogUtil.getProcessDataValueForDeviceElement(context.getIsobusTime().getDLV(), deviceElement, WORK_WIDTH_DDI);
                    break;
                case PROPERTY:
                    workWidth = wwPropertyValues.get(deviceElement);
            }
            if (workWidth == null) {
                return null;
            }
            // distance in meters:
            double distance = context.getGpsDerivedValue().getDistance();
            // affected area in m^2:
            double area = distance * workWidth / 1000; // mm->m
            // return applied amount (uom depends om DDI: mass/volume/count)
            return value * area;
        }

        @Override
        public Double transformOut(AggregationStrategy agg) {
            // sum has total applied amount, return amount/area:
            return agg.getAggregate() / getRasterHelper().getCellArea();
        } 
    }
    
    // TODO: applied rates and yield per time
}
