/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import nl.cloudfarming.client.isobus.model.DDI;
import nl.cloudfarming.client.isobus.model.DeviceElement;
import nl.cloudfarming.client.isobus.model.IsobusDevice;
import nl.cloudfarming.client.isobus.model.IsobusTaskData;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiFileLoader;

/**
 * Data object for the time log xml file which contains the header of the timelog
 * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
 */
public class TimeLogXMLDataObject  extends MultiDataObject {
    
    public static final String MIME_TYPE = "application/x-timelog-xml";
    public static final String ICON_BASE = "";
    private Map<TimeFieldEnum, String> timeMap;
    private Map<PositionFieldEnum, String> positionMap;
    private List<Map<DatalogValueFieldEnum, String>> dataLogValues;
    
    private Map<String, DeviceElement> deviceElements = new HashMap<>();
    private Set<DDI> ddis = new HashSet<>();
    private Set<byte[]> ddiCodes = new HashSet<>();
    private Map<DeviceElement, DeviceElementTemplate> deviceElementDDIMap = new HashMap<>();

    
    protected TimeLogXMLDataObject(FileObject pf, MultiFileLoader loader, List<FileObject> secondary,
            Map<TimeFieldEnum, String> timeMap, Map<PositionFieldEnum, String> positionMap, List<Map<DatalogValueFieldEnum, String>> dataLogValues,
            IsobusTaskData taskData) throws DataObjectExistsException {
        
        super(pf, loader);
        registerEditor(MIME_TYPE, true);
        
        for (FileObject fileObject : secondary) {
            registerEntry(fileObject);
        }

        this.timeMap = timeMap;
        this.positionMap = positionMap;
        this.dataLogValues = dataLogValues;

        initLookups(taskData); 
    }

    private void initLookups(IsobusTaskData taskData) {
        for (Map<DatalogValueFieldEnum, String> template : dataLogValues) {
            String detId = template.get(DatalogValueFieldEnum.DEVICE_ELEMENT_ID_REF);
            String ddiHex = template.get(DatalogValueFieldEnum.PROCESS_DATA_DDI);
            byte[] ddiBytes = DDI.hexToBytes(ddiHex);
            DDI ddi = DDI.findById(ddiBytes);
            
            DeviceElement det =  deviceElements.get(detId);
            if (det == null) {
                det = findDeviceElementByID(detId, taskData);
                deviceElements.put(detId, det);
                deviceElementDDIMap.put(det, new DeviceElementTemplate());
            }
            
            ddiCodes.add(ddiBytes);
            deviceElementDDIMap.get(det).ddiCodes.add(ddiBytes);
            
            if (ddi != null) {
                ddis.add(ddi);
                deviceElementDDIMap.get(det).ddis.add(ddi);
            }
        }
    }

    /**
     * Helper method to find device by ID in the main taskdata file.
     *
     * @param id Id of the device
     * @return Device with the specified ID
     */
    protected DeviceElement findDeviceElementByID(String id, IsobusTaskData isobusTaskData) {
        for (IsobusDevice device : isobusTaskData.getDVC()) {
            for (DeviceElement deviceElement : device.getDET()) {
                if (deviceElement.getId().equals(id)) {
                    return deviceElement;
                }
            }
        }
        return null;
    }
    
    public List<Map<DatalogValueFieldEnum, String>> getDataLogValues() {
        return dataLogValues;
    }

    public Map<PositionFieldEnum, String> getPositionMap() {
        return positionMap;
    }

    public Map<TimeFieldEnum, String> getTimeMap() {
        return timeMap;
    }
    
    public Set<DeviceElement> getDeviceElements() {
        return Collections.unmodifiableSet(deviceElementDDIMap.keySet());
    }

    public DeviceElement findDeviceElementByID(String id) {
        return deviceElements.get(id);
    }
    
    public Set<DDI> getDDIs() {
        return Collections.unmodifiableSet(ddis);
    }
    
    public Set<DDI> getDDIs(DeviceElement det) {
        return Collections.unmodifiableSet(deviceElementDDIMap.get(det).ddis);
    }

    public Set<byte[]> getDDICodes() {
        return Collections.unmodifiableSet(ddiCodes);
    }
    
    public Set<byte[]> getDDICodes(DeviceElement det) {
        return Collections.unmodifiableSet(deviceElementDDIMap.get(det).ddiCodes);
    }

    private class DeviceElementTemplate {
        Set<DDI> ddis = new HashSet<>();
        Set<byte[]> ddiCodes = new HashSet<>();        
    }
}
