/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.type;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * Class to read the header of a binary timeLog file (header is a
 * name-equivalent XML-file)
 *
 * @author Gerben Feenstra & Merijn Zengers
 */
public class TimeLogXMLReader {

    private Document xmlDoc = null;
    private Map<TimeFieldEnum, String> timeMap;
    private Map<PositionFieldEnum, String> positionMap;
    private LinkedList<Map<DatalogValueFieldEnum, String>> dataLogValues;
    private Element timeLogRootElement;

    public TimeLogXMLReader(InputStream xmlInputStream) throws ParserConfigurationException, SAXException, FileNotFoundException {

        DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
        DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
        try {
            xmlDoc = docBuilder.parse(xmlInputStream);
        } catch (IOException ex) {
            throw new FileNotFoundException(ex.getMessage());
        }

        // Get root element (TIM)
        timeLogRootElement = xmlDoc.getDocumentElement();
    }

    /**
     * Gets the Time element from XML Will construct a LinkedHashMap containing
     * a TimeFieldEnum as key and the value from XML as value if the map is
     * already constructed returns the already constructed map
     *
     * @return the constructed map
     */
    Map<TimeFieldEnum, String> getTimeElement() {
        if (timeMap == null) {
            // Keep map ordered by natural ordering
            timeMap = new LinkedHashMap<>();

            for (TimeFieldEnum timeFieldEnum : TimeFieldEnum.values()) {
                // Only attributes that ARE PRESENT are stored in the binary file belonging to the XML
                // So we explicitly check for the existence of the attribute
                if (timeLogRootElement.hasAttribute(timeFieldEnum.getAttributeName())) {
                    String value = timeLogRootElement.getAttribute(timeFieldEnum.getAttributeName());
                    // Empty value will be parsed from the binary file using the binary reader.
                    timeMap.put(timeFieldEnum, value);
                }
            }
        }
        return timeMap;
    }

    /**
     * Gets the Position element from XML Will construct a Linked HashMap
     * containing a PositionFieldEnum as key and the value from the XML as value
     * if the Map is already constructed will return the already constructed map
     *
     * @return the constructed LinkedHashMap
     */
    Map<PositionFieldEnum, String> getPositionElement() {
        if (positionMap == null) {

            positionMap = new LinkedHashMap<>();
            // Get positions list
            NodeList position = timeLogRootElement.getElementsByTagName("PTN");
            Node positionNode;
            Element positionElement = null;
            // Retrieve the Position Element
            if (position.getLength() > 0) {
                positionNode = position.item(0);
                if (positionNode.getNodeType() == Node.ELEMENT_NODE) {
                    positionElement = (Element) positionNode;
                }
            }

            for (PositionFieldEnum fieldEnum : PositionFieldEnum.values()) {
                if (positionElement.hasAttribute(fieldEnum.getAttributeName())) {
                    String value = positionElement.getAttribute(fieldEnum.getAttributeName());
                    positionMap.put(fieldEnum, value);
                }
            }
        }
        return positionMap;
    }

    /**
     * Gets the data log values from the XML Will construct a Linked list
     * containing maps with DatalogValueFieldEnum as key and the value from the
     * XML as value If the linked list is already constructed the list is
     * returned instantly
     *
     * @return The constructed linked list
     */
    List<Map<DatalogValueFieldEnum, String>> getDataLogValues() {
        if (dataLogValues == null) {
            dataLogValues = new LinkedList<>();
            // Get datalogvalues list
            NodeList datalogNodes = timeLogRootElement.getElementsByTagName("DLV");
            Node datalogValueNode;
            Element datalogValue = null;

            for (int i = 0; i < datalogNodes.getLength(); i++) {
                LinkedHashMap dataLogValueMap = new LinkedHashMap<>();
                datalogValueNode = datalogNodes.item(i);
                if (datalogValueNode.getNodeType() == Node.ELEMENT_NODE) {
                    datalogValue = (Element) datalogValueNode;
                }

                for (DatalogValueFieldEnum dataLogEnum : DatalogValueFieldEnum.values()) {
                    if (datalogValue.hasAttribute(dataLogEnum.getAttributeName())) {
                        String value = datalogValue.getAttribute(dataLogEnum.getAttributeName());
                        dataLogValueMap.put(dataLogEnum, value);
                    }
                }
                dataLogValues.add(dataLogValueMap);
            }
        }
        return dataLogValues;
    }
}
